// SIRDS Viewer v0.1, 2005
// Copyright (C) 2005  Giliam de Carpentier
//
// Loads 3D meshes and renders them as Single Image Random Dot Stereograms 
// (SIRDS for short) in real-time.
//
// This program is free software; you can redistribute it and/or modify it under 
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY 
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
// PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with this 
// program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, 
// Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact me at giliam@decarpentier.nl for questions concerning this license.
//
// The source code can be downloaded at http://decarpentier.nl.
//
// This project was written in Microsoft Visual C++ .NET 2003 and is designed to
// work on the Microsoft Windows XP platform. In order to BUILD this project, you 
// need a succesfully installed DirectX 9.0c SDK October 2005 release, which can
// be downloaded from the microsoft website. The project depends on the (included) 
// Microsoft DirectX 9.0c sample framework and the Mersenne Twister implementation 
// coded by Takuji Nishimura and Makoto Matsumoto.
//
// If you are only interested in RUNNING the application, you need a successfully 
// installed version of DirectX 9.0c. 
//
// The application needs to load one or more .X meshes to work. These meshes must 
// be kept in the /Media directory. The sirds_binxp.zip includes a small set of 
// example meshes. These example meshes were not created by me but were found on 
// different internet sites as free downloads and remain the property of their 
// respective owners.
//
//Giliam de Carpentier

#include "dxstdafx.h"
#include "resource.h"
#include "Mersenne/mtrand.h"

//--------------------------------------------------------------------------------------
// Macros
//--------------------------------------------------------------------------------------

// Define some handy macro shortcuts.
#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

// Define maximum number of meshes to be loaded
#define MESH_MAX_POOL					100

// Define n x n (random) dithering buffer size. Must be a power of two.
#define	RANDOM_JITTER_BUFFER_DIM		32

// Define n x n random color buffer (the SIRDS base pattern). Must be a power of two.
#define RANDOM_COLOR_BUFFER_DIM			256

//--------------------------------------------------------------------------------------
// Data structures and enumerations
//--------------------------------------------------------------------------------------

// Define mesh structure.
struct Mesh {
	ID3DXMesh*	pd3dMesh;		// Pointer to a DirectX Mesh
	FLOAT		fRadius;		// Radius bounding sphere
	D3DXVECTOR3 vCenter;		// Center bounding sphere
};

// Enumerate all SIRDS color modes.
enum ColorModel {cmBlackWhite, cmPrimColors, cmSecColors, cmFullGrey, cmFullColors, cmLAST};

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------

// State variables
bool                g_bShowHelp = false;		// If true, it renders the UI control text
bool				g_bUpdateRandomColors = true;// Update SIRDS base random patterne on next render.
bool				g_bEnableSirds = true;		// View grayscale image or SIRDS image
bool				g_bEnableZDithering = true;	// Enable dithering (adds noise to grayscale image)
bool				g_bEnableFreeze = false;	// Re-render SIRDS if no action was detected
bool				g_bCrossEyed = false;		// Stare or cross-eyed mode
ColorModel			g_eCurrentColorModel = cmSecColors; // Current color model

// Ranged paramters and their bounds
int					g_iDepthLevels = 64;		// Current no. of depth levels between back and front plane.
int					g_iDepthLevelsMin = 2;		// Min no. of depth levels
int					g_iDepthLevelsMax = 256;	// Max no. of depth levels

int					g_iRepeatWidth = 119;		// SIRDS base pattern horizontal repeat distance (in pixels)
int					g_iRepeatWidthMin = 20;		// Minimal useful distance between repeating pattern
int					g_iRepeatWidthMax = RANDOM_COLOR_BUFFER_DIM - 1; // Max. useful distance between pattern

int					g_iBlockSize = 2;			// Zoom factor for the random SIRDS base pattern which can 
												// be used to magnify the noise pixels which can make
												// interpretation somewhat more easy.
int					g_iBlockSizeMin = 1;		// Min SIRDS base pattern zoom factor
int					g_iBlockSizeMax = 32;		// Max SIRDS base pattern zoom factor

FLOAT				g_fRampNearPlane = .0f;		// Depth to greylevel near plane
FLOAT				g_fRampNearScale = .4f;		// Depth to greylevel scale

// World/camera data
CModelViewerCamera	g_Camera;					// A model viewing camera
D3DXMATRIXA16       g_mWorldFix;				// World transformation for current mesh

// Mesh data
Mesh		        g_pMeshPool[MESH_MAX_POOL];	// Mesh object
int					g_iMeshPoolSize = 0;		// Number of meshes loaded in the pool
int					g_iMeshActiveIndex = -1;	// Index of mesh in the pool currently active

// DirectX 9 helper structures
ID3DXFont*          g_pFont = NULL;				// Font for drawing text
ID3DXSprite*        g_pSprite = NULL;			// Sprite for batching draw text calls
CDXUTDialogResourceManager g_DialogResourceManager;// manager for shared resources of dialogs
CD3DSettingsDlg     g_SettingsDlg;				// Device settings dialog

// DirectX 9 shader effects
ID3DXEffect*        g_pEffect = NULL;			// D3DX depth-to-greyvalue vertex color conversion shader

// DirectX 9 Surfaces and textures
IDirect3DTexture9*	g_pPostProcessTexture = NULL;// SYSTEMMEM texture for post processing (by the CPU)
												// a copy of the grayvalue rendered mesh on the backbuffer 
IDirect3DSurface9*	g_pPostProcessSurface = NULL;// Handle to post processing texture
IDirect3DSurface9*	g_pBackBufferSurface = NULL;// Handle to back buffer surface

// uint32 random 2D buffers
unsigned long		*g_pRandomColorBuffer = NULL;// 2D color array filled with SIRDS random base pattern
unsigned long		g_pRandomJitterBuffer[RANDOM_JITTER_BUFFER_DIM*RANDOM_JITTER_BUFFER_DIM]; // 2D random
												// array filled with the same noise throughout the run used
												// to jitter the grayvalues before the SIRDS conversion in
												// order to smooth/dither the depth plane transitions.

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           3
#define IDC_CHANGEDEVICE        4


//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
bool    CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext );
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void    CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void    CALLBACK OnLostDevice( void* pUserContext );
void    CALLBACK OnDestroyDevice( void* pUserContext );

void    InitApp();
void	LoadMeshesFromDir( WCHAR* dirName, IDirect3DDevice9* pd3dDevice);
void	ActivateMesh(int newMeshIndex);
void    RenderText( double fTime );
void	postProcessBuffer(char *pBuffer, int width, int height, int bytesLine);
void	postProcessBufferInit(int width, int height);

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // Set the callback functions. These functions allow DXUT to notify
    // the application about device changes, user input, and windows messages.  The 
    // callbacks are optional so you need only set callbacks for events you're interested 
    // in. However, if you don't handle the device reset/lost callbacks then the sample 
    // framework won't be able to reset your device since the application must first 
    // release all device resources before resetting.  Likewise, if you don't handle the 
    // device created/destroyed callbacks then DXUT won't be able to 
    // recreate your device resources.
    DXUTSetCallbackDeviceCreated( OnCreateDevice );
    DXUTSetCallbackDeviceReset( OnResetDevice );
    DXUTSetCallbackDeviceLost( OnLostDevice );
    DXUTSetCallbackDeviceDestroyed( OnDestroyDevice );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( KeyboardProc );
    DXUTSetCallbackFrameRender( OnFrameRender );
    DXUTSetCallbackFrameMove( OnFrameMove );

    // Show the cursor and clip it when in full screen
    DXUTSetCursorSettings( true, true );

	// Initialize the non-framework part of the application
    InitApp();

    // Initialize DXUT and create the desired Win32 window and Direct3D 
    // device for the application. Calling each of these functions is optional, but they
    // allow you to set several options which control the behavior of the framework.
    DXUTInit( true, true, true ); // Parse the command line, handle the default hotkeys, and show msgboxes
    DXUTCreateWindow( L"SIRDS Viewer v0.1, (C) 2005 Giliam de Carpentier. Released under the GNU GPL license." );
    DXUTCreateDevice( D3DADAPTER_DEFAULT, true, 1024, 576, IsDeviceAcceptable, ModifyDeviceSettings );

    // Pass control to DXUT for handling the message pump and 
    // dispatching render calls. DXUT will call your FrameMove 
    // and FrameRender callback when there is idle time between handling window messages.
    DXUTMainLoop();

    // Perform any application-level cleanup here. Direct3D device resources are released within the
    // appropriate callback functions and therefore don't require any cleanup code here.

    return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{

	// Initialize the jitter buffer. This will be left unmodified for the rest of the application run.
	MTRand_int32 mt; // Create a Mersenne Twister random number generator
	int bufferSize = RANDOM_JITTER_BUFFER_DIM * RANDOM_JITTER_BUFFER_DIM;
	for (int i = 0; i < bufferSize; i++) {
		g_pRandomJitterBuffer[i] = mt() % 0xFF; // Store only the lowest byte of the random number
	}

    // Initialize dialogs
    g_SettingsDlg.Init( &g_DialogResourceManager );

}


//--------------------------------------------------------------------------------------
// Called during device initialization, this code checks the device for some 
// minimum set of capabilities, and rejects those that don't pass by returning E_FAIL.
//--------------------------------------------------------------------------------------
bool CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, 
                                  D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    // No fallback defined by this app, so reject any device that 
    // doesn't support at least ps1.1
    if( pCaps->PixelShaderVersion < D3DPS_VERSION(1,1) )
        return false;

	// Do not accept backbuffers with less than 8 bits per color channel as this would 
	// visibly limit for the number of depth levels. Also, limit the backbuffer format
	// to a total of 4 bytes per pixel as this simplifies the CPU post-processing.
	if (BackBufferFormat != D3DFMT_X8R8G8B8 && BackBufferFormat != D3DFMT_A8R8G8B8) return false;

    // Skip backbuffer formats that don't support alpha blending
    IDirect3D9* pD3D = DXUTGetD3DObject(); 
    if( FAILED( pD3D->CheckDeviceFormat( pCaps->AdapterOrdinal, pCaps->DeviceType,
                    AdapterFormat, D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING, 
                    D3DRTYPE_TEXTURE, BackBufferFormat ) ) )
        return false;

	// Otherwise, accept the device
    return true;
}


//--------------------------------------------------------------------------------------
// This callback function is called immediately before a device is created to allow the 
// application to modify the device settings. The supplied pDeviceSettings parameter 
// contains the settings that the framework has selected for the new device, and the 
// application can make any desired changes directly to this structure.  Note however that 
// DXUT will not correct invalid device settings so care must be taken 
// to return valid device settings, otherwise IDirect3D9::CreateDevice() will fail.  
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext )
{
    // If device doesn't support HW T&L or doesn't support 1.1 vertex shaders in HW 
    // then switch to SWVP.
    if( (pCaps->DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) == 0 ||
         pCaps->VertexShaderVersion < D3DVS_VERSION(1,1) )
    {
        pDeviceSettings->BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    }

    // Debugging vertex shaders requires either REF or software vertex processing 
    // and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
    if( pDeviceSettings->DeviceType != D3DDEVTYPE_REF )
    {
        pDeviceSettings->BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
        pDeviceSettings->BehaviorFlags &= ~D3DCREATE_PUREDEVICE;                            
        pDeviceSettings->BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    }
#endif
#ifdef DEBUG_PS
    pDeviceSettings->DeviceType = D3DDEVTYPE_REF;
#endif
    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( pDeviceSettings->DeviceType == D3DDEVTYPE_REF )
            DXUTDisplaySwitchingToREFWarning();
    }

    return true;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// created, which will happen during application initialization and windowed/full screen 
// toggles. This is the best location to create D3DPOOL_MANAGED resources since these 
// resources need to be reloaded whenever the device is destroyed. Resources created  
// here should be released in the OnDestroyDevice callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

	// Initialize dialogs
    V_RETURN( g_DialogResourceManager.OnCreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnCreateDevice( pd3dDevice ) );

    // Initialize the font
    V_RETURN( D3DXCreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
                              OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
                              L"Arial", &g_pFont ) );

	// Reset mesh pool
	g_iMeshActiveIndex = -1;
	g_iMeshPoolSize = 0;

    // Try loading all meshes found in different directories.
	LoadMeshesFromDir(L"./", pd3dDevice);
	LoadMeshesFromDir(L"../", pd3dDevice);
	LoadMeshesFromDir(L"Media/", pd3dDevice);
	LoadMeshesFromDir(L"../Media/", pd3dDevice);

	// Exit if no mesh files were found
	if (FAILED(g_iMeshActiveIndex)) {return -1;};

	ActivateMesh(g_iMeshActiveIndex);

    V_RETURN( CDXUTDirectionWidget::StaticOnCreateDevice( pd3dDevice ) );

    // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the 
    // shader debugger. Debugging vertex shaders requires either REF or software vertex 
    // processing, and debugging pixel shaders requires REF.  The 
    // D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug experience in the 
    // shader debugger.  It enables source level debugging, prevents instruction 
    // reordering, prevents dead code elimination, and forces the compiler to compile 
    // against the next higher available software target, which ensures that the 
    // unoptimized shaders do not exceed the shader model limitations.  Setting these 
    // flags will cause slower rendering since the shaders will be unoptimized and 
    // forced into software.  See the DirectX documentation for more information about 
    // using the shader debugger.
    DWORD dwShaderFlags = D3DXFX_NOT_CLONEABLE;
    #ifdef DEBUG_VS
        dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
    #endif
    #ifdef DEBUG_PS
        dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
    #endif

    // Read the D3DX effect file. sirds.fx contains a vertex shader that sets the
	// color at vertex edges to a depth-correlated greyvalue.
    WCHAR str[MAX_PATH];
    V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"sirds.fx" ) );

    // If this fails, there should be debug output as to 
    // why the .fx file failed to compile
    V_RETURN( D3DXCreateEffectFromFile( pd3dDevice, str, NULL, NULL, dwShaderFlags, NULL, &g_pEffect, NULL ) );

    // Setup the camera's view parameters
    D3DXVECTOR3 vecEye(0.0f, 0.0f, 0.0f);
    D3DXVECTOR3 vecAt (-5.0f, -2.0f, 3.0f);
    g_Camera.SetViewParams( &vecEye, &vecAt );
    g_Camera.SetRadius(2.5f, 1.5f, 10.f );

	// If there wasn't already an SIRDS random color pattern buffer allocated, allocate it now.
	if (!g_pRandomColorBuffer) {
		int bufferSize = RANDOM_COLOR_BUFFER_DIM * RANDOM_COLOR_BUFFER_DIM;
		g_pRandomColorBuffer = new unsigned long[bufferSize];
	}

    return S_OK;
}

//--------------------------------------------------------------------------------------
// This function loads the mesh and ensures the mesh has normals; it also optimizes the 
// mesh for the graphics card's vertex cache, which improves performance by organizing 
// the internal triangle list for less cache misses.
//--------------------------------------------------------------------------------------
void LoadMeshesFromDir(WCHAR* strDirName, IDirect3DDevice9* pd3dDevice)
{
    ID3DXMesh* pMesh = NULL;
	WCHAR	str[MAX_PATH];
    HRESULT hr;

	HANDLE hFindFile;
	WIN32_FIND_DATA FindFileData;

	// Create a wildcard search pattern for the given directory.
	StringCchCopy( str, MAX_PATH, strDirName );
	StringCchCat( str, MAX_PATH, L"*.x" );
	
	// Find the first file matching the pattern
	hFindFile = FindFirstFile(str, &FindFileData);
	if (INVALID_HANDLE_VALUE == hFindFile) return;

	BOOL bFindMoreFiles = true;

	do {

		// Load the mesh with D3DX and get back a ID3DXMesh*.  
		// We'll ignore the X file's embedded materials since materials
		// are not needed to render the depth as a grayvalue.

		// Create a path/filename string
		StringCchCopy( str, MAX_PATH, strDirName );
		StringCchCat( str, MAX_PATH, FindFileData.cFileName );

		// Try to load a mesh from this file
		if (SUCCEEDED(D3DXLoadMeshFromX(str, D3DXMESH_MANAGED, pd3dDevice, NULL, NULL, NULL, NULL, &pMesh))) {

			// Optimize the mesh for this graphics card's vertex cache 
			// so when rendering the mesh's triangle list the vertices will 
			// cache hit more often so it won't have to re-execute the vertex shader 
			// on those vertices so it will improve perf.     
			DWORD *rgdwAdjacency = NULL;
			rgdwAdjacency = new DWORD[pMesh->GetNumFaces() * 3];
			if( rgdwAdjacency == NULL ) {return;}
			V( pMesh->GenerateAdjacency(1e-6f,rgdwAdjacency) );
			V( pMesh->OptimizeInplace(D3DXMESHOPT_VERTEXCACHE, rgdwAdjacency, NULL, NULL, NULL) );
			delete []rgdwAdjacency;

			// Calculate the mesh's bounding sphere radius and center
			D3DXVECTOR3* pData; 
			D3DXVECTOR3 vCenter;
			FLOAT fObjectRadius;

			V( pMesh->LockVertexBuffer( 0, (LPVOID*) &pData ) );
			V( D3DXComputeBoundingSphere( pData, pMesh->GetNumVertices(), D3DXGetFVFVertexSize( pMesh->GetFVF() ), &vCenter, &fObjectRadius ) );
			V( pMesh->UnlockVertexBuffer() );

			// Fill a Mesh structure with the data and add it to the mesh pool.
			Mesh mesh;
			mesh.pd3dMesh = pMesh;
			mesh.vCenter = vCenter;
			mesh.fRadius = fObjectRadius;
			g_pMeshPool[g_iMeshPoolSize++] = mesh;
		}

		// Find the next file matching the file pattern
		bFindMoreFiles = FindNextFile(hFindFile, &FindFileData);

	// If another file matching the pattern was found and the pool isn't saturated yet, process the file
	} while (bFindMoreFiles && g_iMeshPoolSize < MESH_MAX_POOL);

	// Close the file search handle
	FindClose(hFindFile);

	// If one or more meshes were loaded, make the first mesh the active one. -1 otherwise.
	if (g_iMeshPoolSize > 0) {g_iMeshActiveIndex = 0;} else {g_iMeshActiveIndex = -1;}

    return;
}

//--------------------------------------------------------------------------------------
// This function sets the active mesh (i.e. the mesh rendered).
//--------------------------------------------------------------------------------------
void ActivateMesh(int newMeshIndex) {

	// Update state
	g_iMeshActiveIndex = newMeshIndex;

	// Update world transformation (translation and scaling)
	Mesh pMesh = g_pMeshPool[g_iMeshActiveIndex];
    D3DXMatrixTranslation( &g_mWorldFix, -pMesh.vCenter.x, -pMesh.vCenter.y, -pMesh.vCenter.z );
    D3DXMATRIXA16 m;
	FLOAT fScaling = 1.f / pMesh.fRadius;
	D3DXMatrixScaling(&m, fScaling, fScaling, fScaling);
    g_mWorldFix *= m;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// reset, which will happen after a lost device scenario. This is the best location to 
// create D3DPOOL_DEFAULT resources since these resources need to be reloaded whenever 
// the device is lost. Resources created here should be released in the OnLostDevice 
// callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, 
                                const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

	// Get the surface of the backbuffer and cache it
	V(pd3dDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &g_pBackBufferSurface));

	// Create the post processing buffer texture of the same size and format as the backbuffer.
	V_RETURN (
		pd3dDevice->CreateTexture(
			pBackBufferSurfaceDesc->Width,			// width
			pBackBufferSurfaceDesc->Height,			// height
			1,										// levels
			D3DUSAGE_DYNAMIC,						// usage
			pBackBufferSurfaceDesc->Format,			// format
			D3DPOOL_SYSTEMMEM,						// pool
			&g_pPostProcessTexture,					// texture
			NULL) );
	V(g_pPostProcessTexture->GetSurfaceLevel(0, &g_pPostProcessSurface));

	// Let the dialogs know.
    V_RETURN( g_DialogResourceManager.OnResetDevice() );
    V_RETURN( g_SettingsDlg.OnResetDevice() );

	// Let the DX helper structures know
    if( g_pFont )
        V_RETURN( g_pFont->OnResetDevice() );
    if( g_pEffect )
        V_RETURN( g_pEffect->OnResetDevice() );
    V_RETURN( D3DXCreateSprite( pd3dDevice, &g_pSprite ) );

    // Setup the camera's projection parameters
    float fAspectRatio = pBackBufferSurfaceDesc->Width / (FLOAT)pBackBufferSurfaceDesc->Height;
    g_Camera.SetProjParams( D3DX_PI/4, fAspectRatio, 0.1f, 5000.0f );
    g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
    g_Camera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, MOUSE_MIDDLE_BUTTON );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called once at the beginning of every frame. This is the
// best location for your application to handle updates to the scene, but is not 
// intended to contain actual rendering calls, which should instead be placed in the 
// OnFrameRender callback.  
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // Update the camera's position based on user input 
    g_Camera.FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// This callback function will be called at the end of every frame to perform all the 
// rendering calls for the scene, and it will also be called if the window needs to be 
// repainted. After this function has returned, DXUT will call 
// IDirect3DDevice9::Present to display the contents of the next buffer in the swap chain
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // If the settings dialog is being shown, then
    // render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    HRESULT hr;
    D3DXMATRIXA16 mWorldViewProjection;
    UINT iPass, cPasses;
    D3DXMATRIXA16 mWorld;
    D3DXMATRIXA16 mView;
    D3DXMATRIXA16 mProj;

    // Clear the render target and the zbuffer 
    V( pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DXCOLOR(0,0,0,0), 1.0f, 0) );

	// Get back buffer surface description
	D3DSURFACE_DESC BackBufferSurfaceDesc;
	V(g_pBackBufferSurface->GetDesc(&BackBufferSurfaceDesc));

	// Disable back-face culling to minimize model problems
    pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    // Render the scene
    if( SUCCEEDED( pd3dDevice->BeginScene() ) )
    {
        // Get the projection & view matrix from the camera class
        mWorld = g_mWorldFix * *g_Camera.GetWorldMatrix();
        mProj = *g_Camera.GetProjMatrix();
        mView = *g_Camera.GetViewMatrix();

        mWorldViewProjection = mWorld * mView * mProj;

        // Update the effect's variables.  Instead of using strings, it would 
        // be more efficient to cache a handle to the parameter by calling 
        // ID3DXEffect::GetParameterByName
        V( g_pEffect->SetMatrix( "g_mWorldViewProjection", &mWorldViewProjection ) );
        V( g_pEffect->SetFloat( "g_fRampNearPlane", g_fRampNearPlane ) );      
        V( g_pEffect->SetFloat( "g_fRampNearScale", g_fRampNearScale ) );      

        // Apply the shader technique contained in the effect 
        V( g_pEffect->Begin(&cPasses, 0) );

		DWORD dwAttribTableSize;

		ID3DXMesh* pMesh = g_pMeshPool[g_iMeshActiveIndex].pd3dMesh;
		
		pMesh->GetAttributeTable(NULL, &dwAttribTableSize);

		// Render the mesh's passes
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            V( g_pEffect->BeginPass(iPass) );

			// Render the pass with the applied technique
			for (DWORD iAttrib = 0; iAttrib < dwAttribTableSize; iAttrib++) {
				V( pMesh->DrawSubset(iAttrib) );
			}

            V( g_pEffect->EndPass() );
        }
        V( g_pEffect->End() );

		// Copy backbuffer (containing the rendered grayscale mesh) to the postprocessing buffer
		V( pd3dDevice->GetRenderTargetData(g_pBackBufferSurface, g_pPostProcessSurface));

		// Lock the postprocessing buffer
		D3DLOCKED_RECT Rect;
		V( g_pPostProcessSurface->LockRect(&Rect, NULL, 0));

		// Post process (doing the actual SIRDS conversion)
		postProcessBuffer(
			(char *)Rect.pBits, 
			BackBufferSurfaceDesc.Width,
			BackBufferSurfaceDesc.Height,
			Rect.Pitch);

		// Unlock the postprocessing buffer
		g_pPostProcessTexture->UnlockRect(0);

		// Copy the postprocessingbuffer back to the backbuffer    
		pd3dDevice->UpdateSurface(g_pPostProcessSurface, NULL, g_pBackBufferSurface, NULL);

		// Render the text overlay to the backbuffer
        RenderText( fTime );
        
        V( pd3dDevice->EndScene() );
    }
}

//--------------------------------------------------------------------------------------
// Convert the grayscale depth information to a SIRDS image
//--------------------------------------------------------------------------------------
void postProcessBuffer(char *pBuffer, int width, int height, int bytesLine) {

	// Create an offset mask for the jitter buffer
	// Setting this mask to 0 (in case of disabled ZDithering) effectively disables the jitter
	// Setting this mask to RANDOM_JITTER_BUFFER_DIM - 1 enables the maximum mask size
	int jitterMask = (g_bEnableZDithering ? RANDOM_JITTER_BUFFER_DIM - 1 : 0);

	if (!g_bEnableSirds) {
		// SIRDS conversion is currently disabled by the user: Show the greyscale image instead.

		// Move the original greylevel image somewhat to compensate for the SIRDS eye shift.
		int shift = (g_iRepeatWidth - g_iDepthLevels) / 2;
		unsigned long *pImageLine = (unsigned long *)pBuffer;

		if (shift >= 0) {

			// Shift image to the left if base pattern is larger no. of depth levels
			for (int y = 0; y < height; y++) {

				unsigned long *pImage = (unsigned long *)pImageLine;
				int x;
				for (x = 0; x < width - shift; x++, pImage++ ) {*pImage = pImage[shift];}
				for (; x < width; x++, pImage++ ) {*pImage = 0;}

				pImageLine = (unsigned long *)((char *)pImageLine + bytesLine);
			}
		} else {
			// Shift image to the right if the base pattern is smaller than the no. of depth levels
			for (int y = 0; y < height; y++) {

				unsigned long *pImage = (unsigned long *)pImageLine + (width - 1);
				int x;
				for (x = width-1; x > -shift; x--, pImage-- ) {*pImage = pImage[shift];}
				for (; x >= 0; x--, pImage-- ) {*pImage = 0;}

				pImageLine = (unsigned long *)((char *)pImageLine + bytesLine);
			}
		}


	} else {
		// User chose SIRDS rendering mode, so convert the grayscale image to a SIRDS image.

		// Only create a new (random) SIRDS image if it should be (always update if not in 
		// freeze-mode, or only on parameter change and mouse movement when in freeze-mode).
		if (!g_bEnableFreeze) {g_bUpdateRandomColors = true;}

		// Update random color buffer if necessary.
		if (g_bUpdateRandomColors) {

			MTRand_int32 mt; // Create a Mersenne Twister random number generator.
			int bufferSize = RANDOM_COLOR_BUFFER_DIM * RANDOM_COLOR_BUFFER_DIM;
			// Update each color pixel in the random color buffer
			for (int i = 0; i < bufferSize; i++) {
				// Create a random color
				int rnd = mt();
				if (g_eCurrentColorModel == cmPrimColors) {
					// Bound to primary colors if this mode is currently selected.
					rnd &= 0x3;
					rnd = 0xFF << (8 * rnd);
				} else if (g_eCurrentColorModel == cmSecColors) {
					// Bound to secundary colors if this mode is currently selected.
					rnd &= 0x00808080;
					rnd |= (rnd >> 1) | (rnd >> 2);
				} else if (g_eCurrentColorModel == cmBlackWhite) {
					// Bound to black&white if this mode is currently selected.
					rnd = (rnd & 1 ? 0xFFFFFF : 0x000000); 
				} else if (g_eCurrentColorModel == cmFullColors) {
					// Do not bound if full color mode is currently selected.
				} else if (g_eCurrentColorModel == cmFullGrey) {
					// Bound to shades of grey if this mode is currently selected.
					rnd &= 0xFF;
					rnd |= (rnd << 8) | (rnd << 16);
				}
				// Copy the (bounded) color to the buffer pixel
				g_pRandomColorBuffer[i] = 0xFF000000 | rnd;
			}

			g_bUpdateRandomColors = false;
		}

		unsigned long *pImageLine = (unsigned long *)pBuffer;
		int randMask = RANDOM_COLOR_BUFFER_DIM - 1;

		// Convert each greyvalue scanline to a sirds scanline
		for (int y = 0; y < height - 10; y++) {
			// Get starting offset (for most left screen pixel) in random color buffer for this scanline, 
			// wrapping around if necessary.
			unsigned long *pRandLine = g_pRandomColorBuffer + 
				((y / g_iBlockSize) % RANDOM_COLOR_BUFFER_DIM) * RANDOM_COLOR_BUFFER_DIM;

			// Get starting offset (for most left screen pixel) in jitter buffer for this scanline, wrapping
			// around if necessary.
			unsigned long *pJitterLine = g_pRandomJitterBuffer + (y & jitterMask) * RANDOM_JITTER_BUFFER_DIM;

			// Get starting offset (For most top-left screen pixel) in the image buffer
			unsigned long *pImage = (unsigned long *)pImageLine;

			// set filtering value to be XORed with the depth value.
			// setting it to 0 will keep the depth values unaltered.
			// setting it to -1 (which is equal to 0xFFFFFFFF), will reverse the depth, adapting the depth
			// to cross-eyed interpretation.
			int depthFilter = (g_bCrossEyed ? -1 : 0);

			// Create a 16-bit fixed-point fraction increment into the random color buffer for each next screen pixel,
			// creating a pixel zooming effect. Keeping it an integer helps speed up calculations. 
			int blockSizeFract = 0x10000 / g_iBlockSize;

			// Render the SIRDS in two steps. This first step warps the random color image using the depth 
			// (i.e. grayscale) information of the mesh image for the screen rectangle 
			// (0,0) - (g_iRepeatWidth, height - 10), creating the most left strip of the SIRDS image.
			for (int x = 0, randFractX = 0; x < g_iRepeatWidth; x++, pImage++, randFractX += blockSizeFract) {
				// Get depth (here the lowest byte of the color information (being the blue channel from the 
				// grey color), scale it and add some dithering random jitter from the jitter buffer and scale 
				// it again. This creates an additional offset into the current line in the random color buffer.
				int val = (((*pImage ^ depthFilter) & 0xFF) * g_iDepthLevels + pJitterLine[x & jitterMask]) >> 8;
				// Get color from the random color buffer
				*pImage = pRandLine[((randFractX >> 16) + val) & randMask];
			}

			// This second step renders the rest of the SIRDS, warping the SIRDS image to the left using
			// the depth (i.e. grayscale) information.
			for (int x = g_iRepeatWidth; x < width; x++, pImage++ ) {
				// Get depth (here the lowest byte of the color information (being the blue channel from the 
				// grey color), scale it and add some dithering random jitter from the jitter buffer and scale 
				// it again. This creates an additional offset into the SIRDS strip to the left.
				int val = (((*pImage ^ depthFilter) & 0xFF) * g_iDepthLevels + pJitterLine[x & jitterMask]) >> 8;
				// Get color from the SIRDS strip to the left
				*pImage = pImage[val - g_iRepeatWidth];
			}

			// Set starting offset into the mesh image buffer to the next scanline.
			pImageLine = (unsigned long *)((char *)pImageLine + bytesLine);
		}

		// Render the guide dots on the bottom of the buffer (the (0, height-10) - (width, height) rectangle)

		// Calculate position of the two guide dots at the depth of the background
		int xLeft = (width - g_iRepeatWidth + (g_bCrossEyed ? g_iDepthLevels : 0)) / 2;
		int xRight = (width + g_iRepeatWidth - (g_bCrossEyed ? g_iDepthLevels : 0)) / 2;
		// If guide dots are in-screen, render the guide line
		if (g_iRepeatWidth + 10 < width) {
			for (y = height - 10; y < height; y++) {
				unsigned long *pImage = (unsigned long *)pImageLine;
				// Render white region left from left guide dot
				for (int x = 0; x < xLeft-5; x++, pImage++ ) {*pImage = 0xFFFFFFFF;}
				// Render left black guide dot
				for (int x = xLeft-5; x < xLeft+5; x++, pImage++ ) {*pImage = 0xFF000000;}
				// Render white region between left and right guide dot
				for (int x = xLeft+5; x < xRight-5; x++, pImage++ ) {*pImage = 0xFFFFFFFF;}
				// Render right black guide dot
				for (int x = xRight-5; x < xRight+5; x++, pImage++ ) {*pImage = 0xFF000000;}  
				// Render white region right from right guide dot
				for (int x = xRight+5; x < width; x++, pImage++ ) {*pImage = 0xFFFFFFFF;}
				// Set starting offset into the mesh image buffer to the next scanline.
				pImageLine = (unsigned long *)((char *)pImageLine + bytesLine);
			}
		}
	}
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText( double fTime )
{
    // The helper object simply helps keep track of text position, and color
    // and then it calls pFont->DrawText( m_pSprite, strMsg, -1, &rc, DT_NOCLIP, m_clr );
    // If NULL is passed in as the sprite object, then it will work fine however the 
    // pFont->DrawText() will not be batched together.  Batching calls will improves perf.
    CDXUTTextHelper txtHelper( g_pFont, g_pSprite, 15 );

	// Start rendering text
    txtHelper.Begin();
    txtHelper.SetInsertionPos( 2, 0 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
    
    if( g_bShowHelp )
    {
		// The user requested to view the help overlay, so render it.

	    // Draw left column of lines in the help message
        const D3DSURFACE_DESC* pd3dsdBackBuffer = DXUTGetBackBufferSurfaceDesc();
        txtHelper.SetInsertionPos( 2, pd3dsdBackBuffer->Height-15*10 );
        txtHelper.SetForegroundColor( D3DXCOLOR(1.0f, 1.f, 1.0f, 1.0f ) );
        txtHelper.DrawTextLine( L"Controls:" );

        txtHelper.SetInsertionPos( 20, pd3dsdBackBuffer->Height-15*8 );
        txtHelper.DrawFormattedTextLine( L"Hide help: F1\n" 
								L"Settings screen: F2\n"
                                L"Quit: ESC\n"
								L"Rotate model: Left mouse button\n"
                                L"Zoom camera: Mouse wheel scroll\n"
								L"SIRDS conversion on/off: S (%s)\n"
								L"Change model: M / Alt-M\n",
								(g_bEnableSirds ? L"on" : L"off"));

		// Draw right column of lines in the help message
		txtHelper.SetInsertionPos( 320, pd3dsdBackBuffer->Height-15*8 );
		txtHelper.DrawFormattedTextLine(L"Depth levels (back-to-front distance): Q / Alt-Q (%d px)\n"
										L"Repeating pattern (back plane): W / Alt-W (%d px)\n"  
										L"Random color block size: E / Alt-E (%dx%d px)\n"
										L"Color model index: R / Alt-R (%d)\n"
										L"Z-Dithering on/off: D (%s)\n"
										L"Freezing on/off: Space (%s)\n"
										L"Cross-eyed mode: C (%s)\n",
										g_iDepthLevels, 
										g_iRepeatWidth, 
										g_iBlockSize, 
										g_iBlockSize, 
										g_eCurrentColorModel,
										(g_bEnableZDithering ? L"on" : L"off"),
										(g_bEnableFreeze ? L"on" : L"off"),
										(g_bCrossEyed ? L"yes" : L"no"));

    }
    else
    {
		// The user did not request to view the help overlay, so do not render it.
		// Render the 'F1 for help' line instead
        txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
        txtHelper.DrawTextLine( L"Press F1 for help" );
    }
	// Stop rendering text
    txtHelper.End();
}


//--------------------------------------------------------------------------------------
// Before handling window messages, DXUT passes incoming windows 
// messages to the application through this callback function. If the application sets 
// *pbNoFurtherProcessing to TRUE, then DXUT will not process this message.
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Always allow dialog resource manager calls to handle global messages
    // so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

	// An event has taken place. Make sure to update the SIRDS.
	g_bUpdateRandomColors = true;

    // Pass all remaining windows messages to camera so it can respond to user input
    g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}


//--------------------------------------------------------------------------------------
// As a convenience, DXUT inspects the incoming windows messages for
// keystroke messages and decodes the message parameters to pass relevant keyboard
// messages to the application.  The framework does not remove the underlying keystroke 
// messages, which are still passed to the application's MsgProc callback.
//--------------------------------------------------------------------------------------
void CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	// Handle all key presses relevant to the SIRDS rendering
    if( bKeyDown )
    {
        switch( nChar )
        {
            case VK_F1: g_bShowHelp = !g_bShowHelp; break;
            case VK_F2: g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;
            case 'S': g_bEnableSirds = !g_bEnableSirds; break;
			case 'M': ActivateMesh((g_iMeshActiveIndex + g_iMeshPoolSize + (bAltDown?-1:1)) % g_iMeshPoolSize); break;
			case 'Q': g_iDepthLevels = MAX(MIN(g_iDepthLevels + (bAltDown?-1:1), g_iDepthLevelsMax), g_iDepthLevelsMin); break;
			case 'W': g_iRepeatWidth = MAX(MIN(g_iRepeatWidth + (bAltDown?-1:1), g_iRepeatWidthMax), g_iRepeatWidthMin); break;
			case 'E': g_iBlockSize = MAX(MIN(g_iBlockSize + (bAltDown?-1:1), g_iBlockSizeMax), g_iBlockSizeMin); break;
			case 'R': g_eCurrentColorModel = (ColorModel)((g_eCurrentColorModel+(bAltDown?-1:1)+cmLAST)%cmLAST);
					  g_bUpdateRandomColors = true; break;
            case 'D': g_bEnableZDithering = !g_bEnableZDithering; break;
			case 'C': g_bCrossEyed = !g_bCrossEyed; break;
			case ' ': g_bEnableFreeze = !g_bEnableFreeze; break;
        }
    }
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN: DXUTToggleFullScreen(); break;
        case IDC_TOGGLEREF:        DXUTToggleREF(); break;
        case IDC_CHANGEDEVICE:     g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;

    }
    
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// entered a lost state and before IDirect3DDevice9::Reset is called. Resources created
// in the OnResetDevice callback should be released here, which generally includes all 
// D3DPOOL_DEFAULT resources. See the "Lost Devices" section of the documentation for 
// information about lost devices.
//--------------------------------------------------------------------------------------
void CALLBACK OnLostDevice( void* pUserContext )
{
	// Notify dialogs before resetting them
    g_DialogResourceManager.OnLostDevice();
    g_SettingsDlg.OnLostDevice();
    CDXUTDirectionWidget::StaticOnLostDevice();

	// Notify font before resetting it
    if( g_pFont ) g_pFont->OnLostDevice();

	// Notify shader before resetting it
    if( g_pEffect ) g_pEffect->OnLostDevice();

	// Release sprites, textures and surfaces
	SAFE_RELEASE(g_pSprite);
	SAFE_RELEASE(g_pBackBufferSurface);
	SAFE_RELEASE(g_pPostProcessSurface);
    SAFE_RELEASE(g_pPostProcessTexture);
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// been destroyed, which generally happens as a result of application termination or 
// windowed/full screen toggles. Resources created in the OnCreateDevice callback 
// should be released here, which generally includes all D3DPOOL_MANAGED resources. 
//--------------------------------------------------------------------------------------
void CALLBACK OnDestroyDevice( void* pUserContext )
{
	// Notify dialogs of the destruction
    g_DialogResourceManager.OnDestroyDevice();
    g_SettingsDlg.OnDestroyDevice();
    CDXUTDirectionWidget::StaticOnDestroyDevice();

	// Release the shader and font
	SAFE_RELEASE(g_pEffect);
    SAFE_RELEASE(g_pFont);

	// Clean the mesh pool, releasing all meshes contained.
	for (int i = 0; i < g_iMeshPoolSize; i++) {
		SAFE_RELEASE(g_pMeshPool[i].pd3dMesh);
	}
	g_iMeshPoolSize = 0;
	g_iMeshActiveIndex = -1;

	// Delete the SIRDS random color pattern buffer
	SAFE_DELETE(g_pRandomColorBuffer);
}



